import math
import numpy as np
import scipy as scp
from .matrix_solvers import (
    eigen_solvers,
    map_csc_to_eigen,
    translate_sparse,
    cpp_chol,
    compute_Linv,
)
from .penalties import (
    embed_in_S_sparse,
    split_shared_penalties,
    combine_shared_penalties,
)
from .custom_types import LambdaTerm
import copy
import sys


def reparam(
    X: scp.sparse.csc_array | None,
    S: list[LambdaTerm],
    cov: np.ndarray | None,
    option: int = 1,
    n_bins: int = 30,
    QR: bool = False,
    identity: bool = False,
    scale: bool = False,
    form_inverse: int = 0,
    form_root: bool = False,
    n_c: int = 10,
) -> tuple:
    """Options 1 - 3 are natural reparameterization discussed in Wood (2017; 5.4.2)
    with different strategies for the QR computation of :math:`\\mathbf{X}`. Option 4 helps with
    stabilizing the REML computation and is from Appendix B of Wood (2011) and section 6.2.7 in
    Wood (2017):

       1. Form complete matrix :math:`\\mathbf{X}` based on entire covariate.
       2. Form matrix :math:`\\mathbf{X}` only based on unique covariate values.
       3. Form matrix :math:`\\mathbf{X}` on a sample of values making up covariate. Covariate
          is split up into ``n_bins`` equally wide bins. The number of covariate values
          per bin is then calculated. Subsequently, the ratio relative to minimum bin size is
          computed and each ratio is rounded to the nearest integer. Then ``ratio`` samples
          are obtained from each bin. That way, imbalance in the covariate is approximately
          preserved when forming the QR.
       4. Transform term-specific :math:`\\mathbf{S}_{\\boldsymbol{\\lambda}}` based on Appendix B
          of Wood (2011) and section 6.2.7 in Wood (2017) so that they are full-rank and their
          log-determinant can be computed safely. In that case, only ``S`` needs to be provided and
          has to be a list holding the penalties to be transformed. If the transformation is to be
          applied to model matrices, coefficients, hessian, and covariance matrices X should be set
          to something other than ``None`` (does not matter what, can for example be the first model
          matrix.) The :func:`mssm.src.python.gamm_solvers.reparam_model` function can be used to
          apply the transformation and also returns the required transformation matrices to reverse
          it.

    For Options 1-3:

       If ``QR==True`` then :math:`\\mathbf{X}` is decomposed into :math:`\\mathbf{Q}\\mathbf{R}`
       directly via QR decomposition. Alternatively, we first
       form :math:`\\mathbf{X}^T\\mathbf{X}` and then compute the cholesky :math:`\\mathbf{L}` of
       this product - note that :math:`\\mathbf{L}^T = \\mathbf{R}`. Overall the latter
       strategy is much faster (in particular if ``option==1``), but the increased loss of precision
       in :math:`\\mathbf{L}^T = \\mathbf{R}` might not be ok for some.

       After transformation S only contains elements on it's diagonal and :math:`\\mathbf{X}` the
       transformed functions. As discussed in Wood (2017), the transformed functions are
       decreasingly flexible - so the elements on :math:`\\mathbf{S}` diagonal become smaller
       and eventually zero, for elements that are in the kernel of the original
       :math:`\\mathbf{S}` (un-penalized == not flexible).

       For a similar transformation (based solely on :math:`\\mathbf{S}`), Wood et al. (2013) show
       how to further reduce the diagonally transformed :math:`\\mathbf{S}` to an even simpler
       identity penalty. As discussed also in Wood (2017) the same behavior of decreasing
       flexibility if all entries on the diagonal of :math:`\\mathbf{S}` are 1 can only be
       maintained if the transformed functions are multiplied by a weight related to their
       wiggliness. Specifically, more flexible functions need to become smaller in amplitude - so
       that for the same level of penalization they are removed earlier than less flexible ones.
       To achieve this Wood further post-multiply the transformed matrix :math:`\\mathbf{X}'` with
       a matrix that contains on it's diagonal the reciprocal of the square root of the transformed
       penalty matrix (and 1s in the last cells corresponding to the kernel). This is done here
       if ``identity=True``.

       In ``mgcv`` the transformed model matrix and penalty can optionally be scaled by the root
       mean square value of the transformed model matrix (see the nat.param function in mgcv).
       This is done here if ``scale=True``.

    For Option 4:

       Option 4 enforces re-parameterization of term-specific
       :math:`\\mathbf{S}_{\\boldsymbol{\\lambda}}` based on section Wood (2011) and section 6.2.7
       in Wood (2017). In ``mssm`` multiple penalties can be placed on individual terms (i.e.,
       tensor terms, random smooths, Kernel penalty) but
       it is not always the case that the term-specific :math:`\\mathbf{S}_{\\boldsymbol{\\lambda}}`
       - i.e., the sum over all those individual penalties multiplied with
       their :math:`\\lambda` parameters, is of full rank. If we need to form the inverse of the
       term-specific :math:`\\mathbf{S}_{\\boldsymbol{\\lambda}}` this is problematic.
       It is also problematic, as discussed by Wood (2011), if the different :math:`\\lambda` are
       all of different magnitude in which case forming the term-specific
       :math:`log(|\\mathbf{S}_{\\boldsymbol{\\lambda}}|+)` becomes numerically difficult.

       The re-parameterization implemented by option 4, based on Appendix B in Wood (2011), solves
       these issues. After this re-parameterization a term-specific
       :math:`\\mathbf{S}_{\\boldsymbol{\\lambda}}` has been formed that is full rank. And
       :math:`log(|\\mathbf{S}_{\\boldsymbol{\\lambda}}|)` - no longer just a generalized
       determinant - can be computed without running into numerical problems.

       The strategy by Wood (2011) could be applied to form an overall - not just term-specific -
       :math:`\\mathbf{S}_{\\boldsymbol{\\lambda}}` with these properties. However, this
       does not work for general smooth models as defined by Wood et al. (2016). Hence, mssm opts
       for the blockwise strategy. However, in ``mssm`` penalties currently cannot overlap, so this
       is not necessary at the moment.

    References:
     - Wood, S. N., (2011). Fast stable restricted maximum likelihood and marginal likelihood \
        estimation of semiparametric generalized linear models.
     - Wood, S. N., Scheipl, F., & Faraway, J. J. (2013). Straightforward intermediate rank \
        tensor product smoothing in mixed models.
     - Wood, Pya, & Säfken (2016). Smoothing Parameter and Model Selection for General Smooth \
        Models.
     - Wood, S. N. (2017). Generalized Additive Models: An Introduction with R, Second Edition \
        (2nd ed.).
     - mgcv source code (accessed 2024). smooth.R file, nat.param function.

    :param X: Model/Term matrix or None
    :type X: scp.sparse.csc_array | None
    :param S: List of penalties
    :type S: list[LambdaTerm]
    :param cov: covariate array associated with a specific term or None
    :type cov: np.ndarray | None
    :param option: Which re-parameterization to compute, defaults to 1
    :type option: int, optional
    :param n_bins: Number of bins to use as part of option 3, defaults to 30
    :type n_bins: int, optional
    :param QR: Whether to rely on a QR decomposition or not (then a Cholesky is used) as part of
        options 1-3, defaults to False
    :type QR: bool, optional
    :param identity: Whether the penalty matrix should be transformed to identity as part of
        options 1-3, defaults to False
    :type identity: bool, optional
    :param scale: Whether the penalty matrix and term matrix should be scaled as part of
        options 1-3, defaults to False
    :type scale: bool, optional
    :return: Return object content depends on ``option`` but will usually hold informations to
        apply/undo the required re-parameterization as well as already re-parameterized objects.
    :rtype: tuple
    """

    if option < 4:

        # For option 1 just use provided basis matrix
        if option != 1:
            unq, idx, c = np.unique(cov, return_counts=True, return_index=True)
            if option == 2:
                # Form basis based on unique values in cov
                sample = idx
            elif option == 3:
                # Form basis based on re-sampled values of cov to keep row number small but
                # hopefully imbalance
                # in the data preserved.
                weights, values = np.histogram(cov, bins=n_bins)
                ratio = np.round(weights / min(weights), decimals=0).astype(int)

                sample = []
                for bi in range(n_bins - 1):
                    sample_val = np.random.choice(
                        unq[(unq >= values[bi]) & (unq < values[bi + 1])],
                        size=ratio[bi],
                        replace=True,
                    )
                    sample_idx = [
                        idx[unq == sample_val[svi]][0] for svi in range(ratio[bi])
                    ]
                    sample.extend(sample_idx)
                sample.append(idx[-1])
                sample = np.array(sample)

            # Now re-form basis
            X = X[sample, :]

        # Now decompose X = Q @ R
        if QR:
            _, R = scp.linalg.qr(X.toarray(), mode="economic")
            R = scp.sparse.csr_array(R)

        else:
            XX = (X.T @ X).tocsc()

            L, code = eigen_solvers.chol(*map_csc_to_eigen(XX))

            if code != 0:
                raise ValueError("Cholesky failed during reparameterization.")

            R = L.T

        # Now form B and proceed with eigen decomposition of it (see Wood, 2017)
        # see also smooth.R nat.param function in mgcv.
        # R.T @ A = S.T
        # A = Rinv.T @ S.T
        # R.T @ B = A.T
        # A.T = S @ Rinv ## Transpose order reverses!
        # B = Rinv.T @ A.T
        # B = Rinv.T @ S @ Rinv

        B = eigen_solvers.solve_tr(
            *map_csc_to_eigen(R.T),
            eigen_solvers.solve_tr(*map_csc_to_eigen(R.T), S.T).T,
        )

        s, U = scp.linalg.eigh(B.toarray())

        # Decreasing magnitude for ease of indexing..
        s = np.flip(s)
        U = scp.sparse.csc_array(np.flip(U, axis=1))

        rank = len(s[s > 1e-7])

        # First rank elements are non-zero - corresponding to penalized functions,
        # last S.shape[1] - rank are zero corresponding to dimensionality of kernel of S
        Srp = scp.sparse.diags(
            [s[i] if s[i] > 1e-7 else 0 for i in range(S.shape[1])],
            offsets=0,
            format="csc",
        )
        Drp = scp.sparse.diags(
            [s[i] ** 0.5 if s[i] > 1e-7 else 0 for i in range(S.shape[1])],
            offsets=0,
            format="csc",
        )

        # Now compute matrix to transform basis functions. The transformed functions are
        # decreasingly flexible. I.e., Xrp[:,i] is more flexible than Xrp[:,i+1]. According to
        # Wood (2017) Xrp = Q @ U. Now we want to be able to evaluate the basis for new data
        # resulting in Xpred. So we also have to transform Xpred. Following Wood (2017),
        # based on QR decomposition we have X = Q @ R, so we form matrix C so that R @ C = U to
        # have Xrp = Q @ R @ C = Q @ U. Then Xpred_rp = X_pred @ C can similarly be obtained.
        # see smooth.R nat.param function in mgcv.

        C = eigen_solvers.backsolve_tr(*map_csc_to_eigen(R.tocsc()), U)

        IRrp = None
        if identity:
            # Transform S to identity as described in Wood et al. (2013). Form inverse of root of
            # transformed S for all cells not covering a kernel function. For those simply insert 1.
            # Then post-multiply transformed X (or equivalently C) by it.
            IRrp = [1 / s[i] ** 0.5 if s[i] > 1e-7 else 1 for i in range(S.shape[1])]
            Srp = scp.sparse.diags(
                [1 if s[i] > 1e-7 else 0 for i in range(S.shape[1])],
                offsets=0,
                format="csc",
            )
            Drp = copy.deepcopy(Srp)

            C = C @ scp.sparse.diags(IRrp, offsets=0, format="csc")

        rms1 = rms2 = None
        if scale:
            # mgcv optionally scales the transformed model & penalty matrices (separately per range
            # and kernel space columns of S) by the root mean square of the model matrix.
            # see smooth.R nat.param function in mgcv.
            Xrp = X @ C
            rms1 = math.sqrt((Xrp[:, :rank]).power(2).mean())

            # Scale transformation matrix
            C[:, :rank] /= rms1

            # Now apply the separate scaling for Kernel of S as done by mgcv
            if X.shape[1] - rank > 0:
                rms2 = math.sqrt((Xrp[:, rank:]).power(2).mean())
                C[:, rank:] /= rms2

            # Scale penalty
            Srp /= rms1**2
            Drp /= rms1

        # Done, return
        return C, Srp, Drp, IRrp, rms1, rms2, rank

    elif option == 4:
        # Reparameterize S_\\lambda for safe REML evaluation - based on section 6.2.7 in Wood (2017)
        # and Appendix B of Wood (2011). S needs to be list holding penalties.

        # Reparam is best performed based on term-specific penalties rather than lambda-specific
        # penalties
        S = split_shared_penalties(S)

        # We first sort S into term-specific groups of penalties
        SJs = []  # SJs groups per term
        ljs = []  # Lambda groups per term
        SJ_reps = []  # How often should SJ be stacked
        SJ_term_idx = []  # Penalty index of every SJ
        SJ_idx = []  # start index of every SJ
        SJ_coef = []  # Number of coef in the term penalized by a (sum of) SJ
        SJ_idx_max = (
            0  # Max starting index - used to decide whether a new SJ should be added.
        )
        SJ_idx_len = 0  # Number of separate SJ blocks.
        n_coef = S[0].S_J_emb.shape[1] if len(S) > 0 else 0

        # Build S_emb and collect every block for pinv(S_emb)
        for lti, lTerm in enumerate(S):
            # print(f"pen: {lti}")

            # Now collect the S_J for the pinv calculation in the next step.
            if lti == 0 or lTerm.start_index > SJ_idx_max:
                SJs.append([lTerm.S_J])
                ljs.append([lTerm.lam])
                SJ_term_idx.append([lti])
                SJ_reps.append(lTerm.rep_sj)
                SJ_coef.append(lTerm.S_J.shape[1])
                SJ_idx.append(lTerm.start_index)
                SJ_idx_max = lTerm.start_index
                SJ_idx_len += 1

            else:
                # A term with the same starting index exists already - so add the Sjs to the
                # corresponding list
                idx_match = [
                    idx for idx in range(SJ_idx_len) if SJ_idx[idx] == lTerm.start_index
                ]
                # print(idx_match,lTerm.start_index)
                if len(idx_match) > 1:
                    raise ValueError(
                        "Penalty index matches multiple previous locations."
                    )

                SJs[idx_match[0]].append(lTerm.S_J)
                ljs[idx_match[0]].append(lTerm.lam)
                SJ_term_idx[idx_match[0]].append(lti)

                if SJ_reps[idx_match[0]] != lTerm.rep_sj:
                    raise ValueError(
                        "Repeat Number for penalty does not match previous penalties repeat number."
                    )

        # Now we can re-parameterize SJ groups of length > 1
        Sj_reps = []  # Will hold transformed S_J
        Q_reps = []
        for pen in S:
            Sj_reps.append(
                LambdaTerm(
                    S_J=copy.deepcopy(pen.S_J),
                    S_J_emb=copy.deepcopy(pen.S_J_emb),
                    D_J_emb=copy.deepcopy(pen.D_J_emb),
                    rep_sj=pen.rep_sj,
                    lam=pen.lam,
                    type=pen.type,
                    rank=pen.rank,
                    term=pen.term,
                    start_index=pen.start_index,
                    dist_param=pen.dist_param,
                    id=pen.id,
                )
            )

        S_reps = []  # Term specific S_\\lambda
        eps = sys.float_info.epsilon**0.7

        # Number of un-penalized dimensions (Kernel space dimension of total S_\\lambda; Wood, 2011)
        Mp = Sj_reps[0].start_index
        for grp_idx, SJgroup, LJgroup in zip(SJ_term_idx, SJs, ljs):

            for rpidx in range(len(SJgroup)):
                Sj_reps[grp_idx[rpidx]].rp_idx = len(S_reps)

            normed_SJ = [
                SJ_mat / scp.sparse.linalg.norm(SJ_mat, ord="fro") for SJ_mat in SJgroup
            ]
            normed_S = normed_SJ[0]
            for j in range(1, len(normed_SJ)):
                normed_S += normed_SJ[j]

            D, U = scp.linalg.eigh(normed_S.toarray())

            # Decreasing magnitude for ease of indexing..
            D = np.flip(D)
            U = scp.sparse.csc_array(np.flip(U, axis=1))

            Ur = U[:, D > max(D) * eps]
            r = Ur.shape[1]  # Range dimension
            Mp += normed_S.shape[1] - r

            # Initial re-parameterization as discussed by Wood (2011)
            if r < normed_S.shape[1]:
                SJbar = [(Ur.T @ SJ_mat @ Ur).tocsc() for SJ_mat in SJgroup]

                # Need to account for this when computing Q matrix.
                Q_rep0 = copy.deepcopy(U)
            else:
                SJbar = copy.deepcopy(SJgroup)
                Q_rep0 = None

            if len(SJgroup) == 1:  # No further re-parameterization necessary
                Sj_reps[grp_idx[0]].S_J = SJbar[0]
                S_reps.append(SJbar[0] * LJgroup[0])

                if Q_rep0 is not None:
                    Q_reps.append(scp.sparse.csc_array(Q_rep0))
                else:
                    Q_reps.append(scp.sparse.eye(r, format="csc"))

                continue

            S_Jrep = copy.deepcopy(SJbar)  # Original S_J transformed
            S_rep = None
            Q_rep = None

            # Initialization as described by Wood (2011)
            K = 0
            Q = r
            rem_idx = np.arange(0, len(SJbar), dtype=int)

            while True:
                # Find max norm
                norm_group = [
                    scp.sparse.linalg.norm(SJ_mat * SJ_lam, ord="fro")
                    for SJ_mat, SJ_lam in zip(
                        [SJbar[idx] for idx in rem_idx],
                        [LJgroup[idx] for idx in rem_idx],
                    )
                ]

                am_norm = np.argmax(norm_group)

                # Find Sk = S_J with max norm, in Wood (2011) this can be multiple but we
                # just use the max. as discussed in Wood (2017)
                Sk = SJbar[rem_idx[am_norm]] * LJgroup[rem_idx[am_norm]]

                # De-compose Sk - form Ur and Un
                D, U = scp.linalg.eigh(Sk.toarray())

                # Decreasing magnitude for ease of indexing..
                D = np.flip(D)
                U = scp.sparse.csc_array(np.flip(U, axis=1))

                # r is known in advance here almost every-time but we need to de-compose anyway..
                r = len(D[D > max(D) * eps])

                if r == Q:
                    break

                # Seperate U into kernel and rank space of Sk
                n = U.shape[1] - r
                Un = U[:, r:]
                Un = Un.reshape(Sk.shape[1], -1)

                Ur = U[:, :r]
                Ur = Ur.reshape(Sk.shape[1], -1)

                Dr = np.diag(D[:r])

                # Sum up Sjk - remaining S_J not Sk
                Sjk = None
                for j in rem_idx:
                    if j == rem_idx[am_norm]:
                        continue

                    if Sjk is None:
                        Sjk = SJbar[j] * LJgroup[j]
                    else:
                        Sjk += SJbar[j] * LJgroup[j]

                if Sjk is not None:
                    Sjk = Sjk.toarray()
                else:
                    raise ValueError("Problem during re-parameterization.")

                # Compute C' in Wood (2011)

                # Transform Sk
                Sk = Dr + Ur.T @ Sjk @ Ur

                # Fill C' in Wood (2011)
                # print(Sk.shape,(Ur.T@Sjk@Un).shape,(Un.T@Sjk@Ur).shape,(Un.T@Sjk@Un).shape)

                C = np.concatenate(
                    (
                        np.concatenate((Sk, Un.T @ Sjk @ Ur), axis=0),
                        np.concatenate((Ur.T @ Sjk @ Un, Un.T @ Sjk @ Un), axis=0),
                    ),
                    axis=1,
                )

                # print(C.shape)

                # Form S' in Wood (2011)
                if S_rep is None:
                    S_rep = C
                    Ta = np.concatenate(
                        (Ur.toarray(), np.zeros((Ur.shape[0], n))), axis=1
                    )
                    Tg = U.toarray()

                    if Q_rep0 is None:
                        Q_rep = U.toarray()
                    else:
                        init_drop = Q_rep0.shape[1] - U.shape[1]

                        Q_rep = np.concatenate(
                            (
                                np.concatenate(
                                    (U.toarray(), np.zeros((U.shape[0], init_drop))),
                                    axis=1,
                                ),
                                np.concatenate(
                                    (
                                        np.zeros((init_drop, U.shape[1])),
                                        np.identity(init_drop),
                                    ),
                                    axis=1,
                                ),
                            ),
                            axis=0,
                        )
                else:
                    A = S_rep[:K, :K]  # From partitioning step 6 in Wood, 2011
                    B = S_rep[:K, K:]  # From partitioning step 6 in Wood, 2011
                    BU = B @ U
                    S_rep = np.concatenate(
                        (
                            np.concatenate((A, BU.T), axis=0),
                            np.concatenate((BU, C), axis=0),
                        ),
                        axis=1,
                    )

                    Ta = np.concatenate(
                        (
                            np.concatenate(
                                (np.identity(K), np.zeros((K, r + n))), axis=1
                            ),
                            np.concatenate(
                                (
                                    np.zeros((Ur.shape[0], K)),
                                    Ur.toarray(),
                                    np.zeros((Ur.shape[0], n)),
                                ),
                                axis=1,
                            ),
                        ),
                        axis=0,
                    )

                    Tg = np.concatenate(
                        (
                            np.concatenate(
                                (np.identity(K), np.zeros((K, r + n))), axis=1
                            ),
                            np.concatenate(
                                (np.zeros((U.shape[0], K)), U.toarray()), axis=1
                            ),
                        ),
                        axis=0,
                    )

                    if Q_rep0 is None:
                        Q_rep = Q_rep @ Tg
                    else:
                        init_drop = Q_rep0.shape[1] - Tg.shape[1]
                        Q_rep = Q_rep @ np.concatenate(
                            (
                                np.concatenate(
                                    (Tg, np.zeros((Tg.shape[0], init_drop))), axis=1
                                ),
                                np.concatenate(
                                    (
                                        np.zeros((init_drop, Tg.shape[1])),
                                        np.identity(init_drop),
                                    ),
                                    axis=1,
                                ),
                            ),
                            axis=0,
                        )

                # print(Ta.shape,Tg.shape)
                # Transform remaining terms that made up Sjk
                for j in rem_idx:
                    if j == rem_idx[am_norm]:
                        S_Jrep[j] = Ta.T @ S_Jrep[j] @ Ta
                        continue
                    SJbar[j] = scp.sparse.csc_array(Un.T @ SJbar[j] @ Un)
                    S_Jrep[j] = Tg.T @ S_Jrep[j] @ Tg

                K += r
                Q -= r
                rem_idx = np.delete(rem_idx, am_norm)

            for j in range(len(grp_idx)):

                Sj_reps[grp_idx[j]].S_J = scp.sparse.csc_array(S_Jrep[j])

            if S_rep is None:  # Fix r==Q in first iteration
                S_rep = SJbar[0] * LJgroup[0]
                Q_rep = scp.sparse.eye(r, format="csc")

            S_reps.append(scp.sparse.csc_array(S_rep))

            if Q_rep0 is not None:
                Q_rep = scp.sparse.csc_array(Q_rep0 @ Q_rep)
            else:
                Q_rep = scp.sparse.csc_array(Q_rep)

            Q_reps.append(Q_rep)

        # Can now update re-parameterized penalties
        S_emb_rp = None  # Overall transformed penalty
        S_inv_rp = None  # Inverse of overall penalty
        S_root_rp = None  # root of overall penalty

        for Si, (Q_rep, S_rep, grp_idx, S_coef) in enumerate(
            zip(Q_reps, S_reps, SJ_term_idx, SJ_coef)
        ):

            for j in range(len(grp_idx)):
                S_J = Sj_reps[grp_idx[j]].S_J

                # Update embedded S_J
                S_J_emb_rp, _ = embed_in_S_sparse(
                    *translate_sparse(S_J), None, n_coef, S_coef, SJ_idx[Si]
                )
                Q_J_emb, c_idx_j = embed_in_S_sparse(
                    *translate_sparse(Q_rep), None, n_coef, S_coef, SJ_idx[Si]
                )
                for _ in range(1, SJ_reps[Si]):
                    S_J_emb_rp, _ = embed_in_S_sparse(
                        *translate_sparse(S_J), S_J_emb_rp, n_coef, S_coef, c_idx_j
                    )
                    Q_J_emb, c_idx_j = embed_in_S_sparse(
                        *translate_sparse(Q_rep), Q_J_emb, n_coef, S_coef, c_idx_j
                    )

                Sj_reps[grp_idx[j]].S_J_emb = S_J_emb_rp

                # And root
                Sj_reps[grp_idx[j]].D_J_emb = (
                    Q_J_emb.T @ Sj_reps[grp_idx[j]].D_J_emb
                ).tocsc()

                # Also pad S_J (not embedded) with zeros so that shapes remain consistent
                # compared to original parameterization
                S_J_rp, _ = embed_in_S_sparse(
                    *translate_sparse(S_J), None, S_coef, S_coef, 0
                )
                Sj_reps[grp_idx[j]].S_J = S_J_rp

            # Update total penalty and optionally root and inverse

            # Compute inverse of S_rep
            if (
                ((form_inverse == 1) and (len(grp_idx) > 1))
                or form_root
                or form_inverse == 2
            ):
                L, code = cpp_chol(S_rep)
                if code != 0:
                    raise ValueError(
                        "Inverse of transformed penalty could not be computed."
                    )

                S_root_rp, _ = embed_in_S_sparse(
                    *translate_sparse(L), S_root_rp, n_coef, S_coef, SJ_idx[Si]
                )

            # Form inverse - only if this is not a single penalty term
            if ((form_inverse == 1) and (len(grp_idx) > 1)) or form_inverse == 2:
                Linv = compute_Linv(L, n_c)
                S_inv = (Linv.T @ Linv).tocsc()

                S_inv_rp, _ = embed_in_S_sparse(
                    *translate_sparse(S_inv), S_inv_rp, n_coef, S_coef, SJ_idx[Si]
                )

            # Total penalty
            S_emb_rp, c_idx = embed_in_S_sparse(
                *translate_sparse(S_rep), S_emb_rp, n_coef, S_coef, SJ_idx[Si]
            )

            for _ in range(1, SJ_reps[Si]):
                if (
                    ((form_inverse == 1) and (len(grp_idx) > 1))
                    or form_root
                    or form_inverse == 2
                ):
                    S_root_rp, _ = embed_in_S_sparse(
                        *translate_sparse(L), S_root_rp, n_coef, S_coef, c_idx
                    )

                if ((form_inverse == 1) and (len(grp_idx) > 1)) or form_inverse == 2:
                    S_inv_rp, _ = embed_in_S_sparse(
                        *translate_sparse(S_inv), S_inv_rp, n_coef, S_coef, c_idx
                    )

                S_emb_rp, c_idx = embed_in_S_sparse(
                    *translate_sparse(S_rep), S_emb_rp, n_coef, S_coef, c_idx
                )

        return (
            Sj_reps,
            S_emb_rp,
            S_inv_rp,
            S_root_rp,
            S_reps,
            SJ_term_idx,
            SJ_idx,
            SJ_coef,
            Q_reps,
            Mp,
        )

    else:
        raise NotImplementedError(
            f"Requested option {option} for reparameterization is not implemented."
        )


def reparam_model(
    dist_coef: list[int],
    dist_up_coef: list[int],
    coef: np.ndarray,
    split_coef_idx: list[int],
    Xs: list[scp.sparse.csc_array],
    penalties: list[LambdaTerm],
    form_inverse: bool = True,
    form_root: bool = True,
    form_balanced: bool = True,
    n_c: int = 1,
) -> tuple[
    np.ndarray,
    list[scp.sparse.csc_array],
    list[LambdaTerm],
    scp.sparse.csc_array,
    scp.sparse.csc_array | None,
    scp.sparse.csc_array | None,
    scp.sparse.csc_array | None,
    scp.sparse.csc_array,
    list[scp.sparse.csc_array],
]:
    """Relies on the transformation strategy from Appendix B of Wood (2011) to re-parameterize the
    model.

    Coefficients, model matrices, and penalties are all transformed. The transformation is applied
    to each term separately as explained by Wood et al., (2016).

    References:
     - Wood, Pya, & Säfken (2016). Smoothing Parameter and Model Selection for General Smooth \
        Models.

    :param dist_coef: List of number of coefficients per formula/linear predictor/distribution
        parameter of model.
    :type dist_coef: [int]
    :param dist_up_coef: List of number of **unpenalized** (i.e., fixed effects, linear
        predictors/parameters) coefficients per formula/linear predictor/distribution parameter
        of model.
    :type dist_up_coef: [int]
    :param coef: Vector of coefficients (numpy.array of dim (-1,1)).
    :type coef: numpy.array
    :param split_coef_idx: List with indices to split ``coef`` vector into separate versions per
        linear predictor.
    :type split_coef_idx: [int]
    :param Xs: List of model matrices obtained for example via ``model.get_mmat()``.
    :type Xs: [scp.sparse.csc_array]
    :param penalties: List of penalties for model.
    :type penalties: [LambdaTerm]
    :param form_inverse: Whether or not an inverse of the transformed penalty matrices should be
        formed. Useful for computing the EFS update, defaults to True
    :type form_inverse: bool, optional
    :param form_root: Whether or not to form a root of the total penalty, defaults to True
    :type form_root: bool, optional
    :param form_balanced: Whether or not to form the "balanced" penalty as described by
        Wood et al. (2016) after the re-parameterization, defaults to True
    :type form_balanced: bool, optional
    :param n_c: Number of cores to use to ocmpute the inverse when ``form_inverse=True``,
        defaults to 1
    :type n_c: int, optional
    :raises ValueError: Raises a value error if one of the inverse computations fails.
    :return: A tuple with 9 elements: the re-parameterized coefficient vector, a list with the
        re-parameterized model matrices, a list of the penalties after re-parameterization, the
        total re-parameterized penalty matrix, optionally the balanced version of the former,
        optionally a root of the re-parameterized total penalty matrix, optionally the inverse of
        the re-parameterized total penalty matrix, the transformation matrix ``Q`` so that
        ``Q.T@S_emb@Q = S_emb_rp`` where ``S_emb`` and ``S_emb_rp`` are the total penalty matrix
        before and after re-parameterization, a list of transformation matrices ``QD`` so that
        ``XD@QD=XD_rp`` where ``XD`` and ``XD_rp`` are the model matrix of the Dth linear predictor
        before and after re-parameterization.
    :rtype: tuple[np.ndarray, list[scp.sparse.csc_array], list[LambdaTerm], scp.sparse.csc_array,
        scp.sparse.csc_array | None, scp.sparse.csc_array | None, scp.sparse.csc_array | None,
        scp.sparse.csc_array, list[scp.sparse.csc_array]]
    """

    # Apply reparam from Wood (2011) Appendix B
    (
        Sj_reps,
        S_emb_rp,
        S_inv_rp,
        S_root_rp,
        S_reps,
        SJ_term_idx,
        _,
        S_coefs,
        Q_reps,
        _,
    ) = reparam(
        Xs[0],
        penalties,
        None,
        option=4,
        form_inverse=int(form_inverse),
        form_root=form_root,
    )

    S_norm_rp = None
    Q_emb = None

    dist_idx = Sj_reps[0].dist_param
    Qs = []

    # Create transformation matrix for first dist. parameter/linear predictor and overall
    Q_emb, _ = embed_in_S_sparse(
        *translate_sparse(scp.sparse.identity(dist_up_coef[0], format="csc")),
        None,
        len(coef),
        dist_up_coef[0],
        0,
    )
    Qd_emb, c_idx = embed_in_S_sparse(
        *translate_sparse(scp.sparse.identity(dist_up_coef[0], format="csc")),
        None,
        dist_coef[0],
        dist_up_coef[0],
        0,
    )
    cd_idx = c_idx

    for Si, (S_rep, S_coef) in enumerate(zip(S_reps, S_coefs)):

        # Create new Q if we move to a new parameter
        if Sj_reps[SJ_term_idx[Si][0]].dist_param != dist_idx:
            dist_idx += 1
            Qs.append(Qd_emb)

            # Make sure to update c_idx and cdidx here
            cd_idx = 0
            Qd_emb, cd_idx = embed_in_S_sparse(
                *translate_sparse(
                    scp.sparse.identity(dist_up_coef[dist_idx], format="csc")
                ),
                None,
                dist_coef[dist_idx],
                dist_up_coef[dist_idx],
                cd_idx,
            )
            Q_emb, c_idx = embed_in_S_sparse(
                *translate_sparse(
                    scp.sparse.identity(dist_up_coef[dist_idx], format="csc")
                ),
                Q_emb,
                len(coef),
                dist_up_coef[dist_idx],
                c_idx,
            )
            # c_idx += dist_up_coef[dist_idx]

        # Continue to fill overall penalty matrix and current Q
        for _ in range(Sj_reps[SJ_term_idx[Si][0]].rep_sj):

            # Transformation matrix for current linear predictor
            Qd_emb, cd_idx = embed_in_S_sparse(
                *translate_sparse(Q_reps[Si]),
                Qd_emb,
                dist_coef[dist_idx],
                S_coef,
                cd_idx,
            )

            # Overall transformation matrix for penalties
            Q_emb, c_idx = embed_in_S_sparse(
                *translate_sparse(Q_reps[Si]), Q_emb, len(coef), S_coef, c_idx
            )

    # Fill remaining cells of Q_emb in case final dist. parameter/linear predictor
    # has only unpenalized coef
    if c_idx != len(coef):
        Q_emb, c_idx = embed_in_S_sparse(
            *translate_sparse(scp.sparse.identity(len(coef) - c_idx, format="csc")),
            Q_emb,
            len(coef),
            len(coef) - c_idx,
            c_idx,
        )

    # Collect final Q
    Qs.append(Qd_emb)

    if form_balanced:
        # Compute balanced penalty in reparam
        S_norm_rp = copy.deepcopy(Sj_reps[0].S_J_emb) / scp.sparse.linalg.norm(
            Sj_reps[0].S_J_emb, ord=None
        )
        for peni in range(1, len(Sj_reps)):
            S_norm_rp += Sj_reps[peni].S_J_emb / scp.sparse.linalg.norm(
                Sj_reps[peni].S_J_emb, ord=None
            )

        S_norm_rp /= scp.sparse.linalg.norm(S_norm_rp, ord=None)

    # Transform model matrices
    Xs_rp = copy.deepcopy(Xs)
    for qi, Q in enumerate(Qs):
        # print(Xs[qi].shape,Q.shape)
        Xs_rp[qi] = Xs[qi] @ Q

    # Transform coef
    coef_rp = copy.deepcopy(coef)
    split_coef_rp = np.split(coef_rp, split_coef_idx)

    for qi, Q in enumerate(Qs):
        split_coef_rp[qi] = Q.T @ split_coef_rp[qi]

    coef_rp = np.concatenate(split_coef_rp).reshape(-1, 1)

    # Re-combine any shared penalties
    Sj_reps = combine_shared_penalties(Sj_reps)

    return coef_rp, Xs_rp, Sj_reps, S_emb_rp, S_norm_rp, S_root_rp, S_inv_rp, Q_emb, Qs



import copy
import numpy as np
import scipy as scp
from .matrix_solvers import cpp_cholP, compute_Linv, apply_eigen_perm

########## Compact Representations of quasi-Newton updates for L-qEFS update ########## # noqa: E266


def compute_omega(yks: np.ndarray, sks: np.ndarray, method: str = "NoWr") -> float:
    """Computes scaling factor for quasi-Newton Hessian approximation.


    Setting ``method = "NoWr"`` computes the scaling factor as described by
    Nocedal & Wright (2004) for the final set of update vectors. Setting ``method = "mean"`` or
    ``method = "min"`` computes the scaling factor as described by Brust (2020) for every pair
    of update vectors and then either returns the mean or minimum.

    References:
     - Nocedal, J., & Wright, S. J. (2006). Numerical Optimization.\
        https://doi.org/10.1007/978-0-387-40065-5
     - Brust, J. (2020) Limited Memory Structured Quasi-Newton Methods. A LANS Seminar Presentation

    :param yks: Array of update vectors yk
    :type yks: np.ndarray
    :param sks: Array of update vectors sk
    :type sks: np.ndarray
    :param method: Which method to use to compute scaling factor. Default is version by Nocedal &
        Wright (2004).
    :type method: str, optional
    :return: Scaling factor for initial approximation to hessian.
    :rtype: float
    """
    if method == "NoWr":
        omega = np.dot(yks[-1], yks[-1]) / np.dot(yks[-1], sks[-1])

    elif method in ["mean", "min"]:
        omega = 0
        omega_min = 0

        for ui in range(yks.shape[0]):
            omegai = np.dot(sks[ui], yks[ui]) / np.dot(sks[ui], sks[ui])
            omega += omegai

            if omegai < omega_min or omega_min <= 0:
                omega_min = omegai

        omega /= yks.shape[0]
        if method == "min":
            omega = omega_min

    if omega < 0:
        omega = 1

    return omega


def computeH(
    s: np.ndarray,
    y: np.ndarray,
    rho: np.ndarray,
    H0: scp.sparse.csc_array,
    explicit: bool = True,
) -> (
    np.ndarray
    | tuple[
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]
):
    """Computes (explicitly or implicitly) the quasi-Newton approximation to the negative Hessian
    of the (penalized) likelihood :math:`\\mathbf{H}` (:math:`\\mathcal{H}`) from the L-BFGS-B
    optimizer info.

    Relies on equations 2.16 in Byrd, Nocdeal & Schnabel (1992).

    References:
     - Byrd, R. H., Nocedal, J., & Schnabel, R. B. (1994). Representations of quasi-Newton \
        matrices and their use in limited memory methods. Mathematical Programming, 63(1), \
        129–156. https://doi.org/10.1007/BF01582063

    :param s: np.ndarray of shape (m,p), where p is the number of coefficients, holding the first
        set ``m`` of update vectors from Byrd, Nocdeal & Schnabel (1992).
    :type s: np.ndarray
    :param y: np.ndarray of shape (m,p), where p is the number of coefficients, holding the second
        set ``m`` of update vectors from Byrd, Nocdeal & Schnabel (1992).
    :type y: np.ndarray
    :param rho: flattened numpy.array of shape (m,), holding element-wise ``1/y.T@s`` from Byrd,
        Nocdeal & Schnabel (1992).
    :type rho: np.ndarray
    :param H0: Initial estimate for the hessian of the negative (penalized) likelihood. Here some
        multiple of the identity (multiplied by ``omega``).
    :type H0: scipy.sparse.csc_array
    :param explicit: Whether or not to return the approximate matrix explicitly or implicitly in
        form of four update matrices.
    :type explicit: bool
    :return: H, either as np.ndarray (``explicit=='True'``) or represented implicitly via four
        update vectors (also np.ndarrays)
    :rtype: np.ndarray | tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
    """
    # Number of updates?
    m = len(y)

    # First form S,Y, and D
    S = s.T
    Y = y.T

    STS = S.T @ H0 @ S
    DK = np.identity(m)
    DK[0, 0] *= np.dot(s[0], y[0])

    # Now use eq. 2.5 to compute R - only have to do this once
    R0 = np.dot(s[0], y[0]).reshape(1, 1)
    R = R0
    for k in range(1, m):

        DK[k, k] *= np.dot(s[k], y[k])

        R = np.concatenate(
            (
                np.concatenate((R0, S[:, :k].T @ Y[:, [k]]), axis=1),
                np.concatenate(
                    (np.zeros((1, R0.shape[1])), np.array([1 / rho[k]]).reshape(1, 1)),
                    axis=1,
                ),
            ),
            axis=0,
        )

        R0 = R

    # Eq 2.22
    L = S.T @ Y - R

    # Now compute term 2 in 3.13 of Byrd, Nocdeal & Schnabel (1992)
    t2 = np.zeros((2 * m, 2 * m))
    t2[:m, :m] = STS
    t2[:m, m:] = L
    t2[m:, :m] = L.T
    t2[m:, m:] = -1 * DK

    # We actually need the inverse to compute H

    # Eq 2.26 of Byrd, Nocdeal & Schnabel (1992)
    Dinv = copy.deepcopy(DK)
    Dpow = copy.deepcopy(DK)
    Dnpow = copy.deepcopy(DK)
    for k in range(m):
        Dinv[k, k] = 1 / Dinv[k, k]
        Dpow[k, k] = np.power(Dpow[k, k], 0.5)
        Dnpow[k, k] = np.power(Dnpow[k, k], -0.5)

    JJT = STS + L @ Dinv @ L.T
    J = scp.linalg.cholesky(JJT, lower=True)

    t2L = np.zeros((2 * m, 2 * m))
    t2L[:m, :m] = Dpow
    t2L[m:, :m] = (-1 * L) @ Dnpow
    t2L[m:, m:] = J

    t2U = np.zeros((2 * m, 2 * m))
    t2U[:m, :m] = -1 * Dpow
    t2U[:m:, m:] = Dnpow @ L.T
    t2U[m:, m:] = J.T

    t2_flip = t2L @ t2U

    invt2L = scp.linalg.inv(t2L)
    invT2U = scp.linalg.inv(t2U)
    invt2 = invt2L.T @ invT2U.T

    t2_sort = np.zeros((2 * m, 2 * m))
    # top left <- bottom right
    t2_sort[:m, :m] = t2_flip[m:, m:]
    # top right <- bottom left
    t2_sort[:m, m:] = t2_flip[m:, :m]
    # bottom left <- top right
    t2_sort[m:, :m] = t2_flip[:m, m:]
    # bottom right <- top left
    t2_sort[m:, m:] = t2_flip[:m, :m]

    invt2_sort = np.zeros((2 * m, 2 * m))
    # top left <- bottom right
    invt2_sort[:m, :m] = invt2[m:, m:]
    # top right <- bottom left
    invt2_sort[:m, m:] = invt2[m:, :m]
    # bottom left <- top right
    invt2_sort[m:, :m] = invt2[:m, m:]
    # bottom right <- top left
    invt2_sort[m:, m:] = invt2[:m, :m]

    # And terms 1 and 2
    t1 = np.concatenate((H0 @ S, Y), axis=1)
    t3 = np.concatenate((S.T @ H0, Y.T), axis=0)

    # Return matrix in compact representation
    if explicit is False:
        return t1, (-1 * t2_sort), (-1 * invt2_sort), t3

    H = H0 + t1 @ (-1 * invt2_sort) @ t3

    return H


def computeV(
    s: np.ndarray,
    y: np.ndarray,
    rho: np.ndarray,
    V0: scp.sparse.csc_array,
    explicit: bool = True,
) -> np.ndarray | tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Computes (explicitly or implicitly) the quasi-Newton approximation to the inverse of the
    negative Hessian of the (penalized) likelihood :math:`\\mathcal{I}` (:math:`\\mathbf{V}`)
    from the L-BFGS-B optimizer info.

    Relies on equations 2.16 and 3.13 in Byrd, Nocdeal & Schnabel (1992).

    References:
     - Byrd, R. H., Nocedal, J., & Schnabel, R. B. (1994). Representations of quasi-Newton \
        matrices and their use in limited memory methods. Mathematical Programming, 63(1), \
        129–156. https://doi.org/10.1007/BF01582063

    :param s: np.ndarray of shape (m,p), where p is the number of coefficients, holding the
        first set ``m`` of update vectors from Byrd, Nocdeal & Schnabel (1992).
    :type s: np.ndarray
    :param y: np.ndarray of shape (m,p), where p is the number of coefficients, holding the second
        set ``m`` of update vectors from Byrd, Nocdeal & Schnabel (1992).
    :type y: np.ndarray
    :param rho: flattened numpy.array of shape (m,), holding element-wise ```1/y.T@s`` from Byrd,
        Nocdeal & Schnabel (1992).
    :type rho: np.ndarray
    :param V0: Initial estimate for the inverse of the hessian of the negative (penalized)
        likelihood. Here some multiple of the identity (multiplied by ``omega``).
    :type V0: scipy.sparse.csc_array
    :param explicit: Whether or not to return the approximate matrix explicitly or implicitly in
        form of three update matrices.
    :type explicit: bool
    :return: V, either as np.ndarray (``explicit=='True'``) or represented implicitly via three
        update vectors (also np.ndarrays)
    :rtype: np.ndarray | tuple[np.ndarray, np.ndarray, np.ndarray]
    """
    m = len(y)
    # First form S,Y, and D
    S = s.T
    Y = y.T

    DYTY = Y.T @ V0 @ Y

    DYTY[0, 0] += np.dot(s[0], y[0])

    # Now use eq. 2.5 to compute R^{-1} - only have to do this once
    Rinv0 = 1 / np.dot(s[0], y[0]).reshape(1, 1)
    Rinv = Rinv0
    for k in range(1, m):

        DYTY[k, k] += np.dot(s[k], y[k])

        Rinv = np.concatenate(
            (
                np.concatenate(
                    (Rinv0, (-rho[k]) * Rinv0 @ S[:, :k].T @ Y[:, [k]]), axis=1
                ),
                np.concatenate(
                    (np.zeros((1, Rinv0.shape[1])), np.array([rho[k]]).reshape(1, 1)),
                    axis=1,
                ),
            ),
            axis=0,
        )

        Rinv0 = Rinv

    # Now compute term 2 in 3.13 used for all S_j
    t2 = np.zeros((2 * m, 2 * m))
    t2[:m, :m] = Rinv.T @ DYTY @ Rinv
    t2[:m, m:] = -Rinv.T
    t2[m:, :m] = -Rinv

    # And terms 1 and 2
    t1 = np.concatenate((S, V0 @ Y), axis=1)
    t3 = np.concatenate((S.T, Y.T @ V0), axis=0)

    if explicit:
        V = V0 + t1 @ t2 @ t3
        return V
    else:
        return t1, t2, t3


def computeVSR1(
    s: np.ndarray,
    y: np.ndarray,
    rho: np.ndarray,
    V0: scp.sparse.csc_array,
    omega: float = 1,
    make_psd: bool = False,
    explicit: bool = True,
) -> np.ndarray | tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Computes (explicitly or implicitly) the symmetric rank one (SR1) approximation of the
    inverse of the negative Hessian of the (penalized) likelihood :math:`\\mathcal{I}`
    (:math:`\\mathbf{V}`).

    Relies on equations 2.16 and 3.13 in Byrd, Nocdeal & Schnabel (1992). Can ensure positive
    (semi) definiteness of the approximation via an eigen decomposition as shown by
    Burdakov et al. (2017). This is enforced via the ``make_psd`` argument.

    References:
     - Burdakov, O., Gong, L., Zikrin, S., & Yuan, Y. (2017). On efficiently combining \
        limited-memory and trust-region techniques. Mathematical Programming Computation, \
        9(1), 101–134. https://doi.org/10.1007/s12532-016-0109-7
     - Byrd, R. H., Nocedal, J., & Schnabel, R. B. (1994). Representations of quasi-Newton \
        matrices and their use in limited memory methods. Mathematical Programming, 63(1), \
        129–156. https://doi.org/10.1007/BF01582063

    :param s: np.ndarray of shape (m,p), where p is the number of coefficients, holding the first
        set ``m`` of update vectors from Byrd, Nocdeal & Schnabel (1992).
    :type s: np.ndarray
    :param y: np.ndarray of shape (m,p), where p is the number of coefficients, holding the second
        set ``m`` of update vectors from Byrd, Nocdeal & Schnabel (1992).
    :type y: np.ndarray
    :param rho: flattened numpy.array of shape (m,), holding element-wise ```1/y.T@s`` from Byrd,
        Nocdeal & Schnabel (1992).
    :type rho: np.ndarray
    :param V0: Initial estimate for the inverse of the hessian of the negative (penalized)
        likelihood. Here some multiple of the identity (multiplied by ``omega``).
    :type V0: scipy.sparse.csc_array
    :param omega: Multiple of the identity matrix used as initial estimate.
    :type omega: float, optional
    :param make_psd: Whether to enforce PSD as mentioned in the description. By default set to
        False.
    :type make_psd: bool, optional
    :param explicit: Whether or not to return the approximate matrix explicitly or implicitly in
        form of three update matrices.
    :type explicit: bool
    :return: V, either as np.ndarray (``explicit=='True'``) or represented implicitly via three
        update vectors (also np.ndarrays)
    :rtype: np.ndarray | tuple[np.ndarray, np.ndarray, np.ndarray]
    """
    m = len(y)
    # First form S,Y, and D
    S = s.T
    Y = y.T

    YTY = Y.T @ V0 @ Y
    DK = np.identity(m)
    DK[0, 0] *= np.dot(s[0], y[0])

    # Now use eq. 2.5 to compute R - only have to do this once
    R0 = np.dot(s[0], y[0]).reshape(1, 1)
    R = R0
    for k in range(1, m):

        DK[k, k] *= np.dot(s[k], y[k])

        R = np.concatenate(
            (
                np.concatenate((R0, S[:, :k].T @ Y[:, [k]]), axis=1),
                np.concatenate(
                    (np.zeros((1, R0.shape[1])), np.array([1 / rho[k]]).reshape(1, 1)),
                    axis=1,
                ),
            ),
            axis=0,
        )

        R0 = R

    # Eq 2.22
    # L = S.T @ Y - R

    # Now compute term 2 in eq. 5.2
    t2 = scp.linalg.inv(R + R.T - DK - YTY)

    # And terms 1 and 2
    t1 = S - V0 @ Y
    t3 = t1.T

    if make_psd:
        # Compute implicit eigen decomposition as shown by Burdakov et al. (2017)
        Q, R = scp.linalg.qr(t1, mode="economic")
        Rit2R = R @ (t2) @ R.T

        # ev holds non-zero eigenvalues of U@D@U.T (e.g., Burdakov et al. (2017))
        ev, P = scp.linalg.eigh(Rit2R, driver="ev")

        # Now find closest PSD.
        fix_idx = (ev + omega) <= 0

        if np.sum(fix_idx) > 0:
            # print("fix VSR1",np.sum(fix_idx),omega,1/omega)
            ev[fix_idx] = -1 * omega  # + np.power(np.finfo(float).eps,0.9)

            # while np.any(np.abs(ev) < 1e-7):
            #   ev[np.abs(ev) < 1e-7] += np.power(np.finfo(float).eps,0.7)

            # print(f"implicit ev post shift. abs min: {np.min(np.abs(ev))}, min: {np.min(ev)},
            #   min + 1: {np.min(ev+omega)}, max + 1: {np.max(ev+omega)}",ev[:10]+omega)

            # Now Q @ P @ np.diag(ev) @ P.T @ Q.T = shifted PSD version of U@D@U.T
            # so we can set:
            # shifted_invt2=np.diag(ev)
            # shifted_t2 = np.diag(1/ev)
            # t1 = Q @ P
            # t3 = t1.T = P.T @ Q.T
            t1 = Q @ P
            t3 = P.T @ Q.T
            t2 = np.diag(ev)

    if explicit:
        V = V0 + t1 @ t2 @ t3
        return V
    else:
        return t1, t2, t3


def computeHSR1(
    s: np.ndarray,
    y: np.ndarray,
    rho: np.ndarray,
    H0: scp.sparse.csc_array,
    omega: float = 1,
    make_psd: bool = False,
    make_pd: bool = False,
    explicit: bool = True,
) -> np.ndarray | tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Computes, (explicitly or implicitly) the symmetric rank one (SR1) approximation of the
    negative Hessian of the (penalized) likelihood :math:`\\mathbf{H}` (:math:`\\mathcal{H}`).

    Relies on equations 2.16 and 3.13 in Byrd, Nocdeal & Schnabel (1992). Can ensure positive (semi)
    definiteness of the approximation via an eigen decomposition as shown by Burdakov et al. (2017).
    This is enforced via the ``make_psd`` and ``make_pd`` arguments.

    References:
     - Burdakov, O., Gong, L., Zikrin, S., & Yuan, Y. (2017). On efficiently combining \
        limited-memory and trust-region techniques. Mathematical Programming Computation, \
        9(1), 101–134. https://doi.org/10.1007/s12532-016-0109-7
     - Byrd, R. H., Nocedal, J., & Schnabel, R. B. (1994). Representations of quasi-Newton \
        matrices and their use in limited memory methods. Mathematical Programming, 63(1), \
        129–156. https://doi.org/10.1007/BF01582063

    :param s: np.ndarray of shape (m,p), where p is the number of coefficients, holding the
        first set ``m`` of update vectors from Byrd, Nocdeal & Schnabel (1992).
    :type s: np.ndarray
    :param y: np.ndarray of shape (m,p), where p is the number of coefficients, holding the
        second set ``m`` of update vectors from Byrd, Nocdeal & Schnabel (1992).
    :type y: np.ndarray
    :param rho: flattened numpy.array of shape (m,), holding element-wise ``1/y.T@s`` from
        Byrd, Nocdeal & Schnabel (1992).
    :type rho: np.ndarray
    :param H0: Initial estimate for the hessian of the negative (penalized) likelihood. Here
        some multiple of the identity (multiplied by ``omega``).
    :type H0: scipy.sparse.csc_array
    :param omega: Multiple of the identity matrix used as initial estimate.
    :type omega: float, optional
    :param make_psd: Whether to enforce PSD as mentioned in the description. By default set to
        False.
    :type make_psd: bool, optional
    :param make_pd: Whether to enforce numeric positive definiteness, not just PSD. Ignored if
        ``make_psd=False``. By default set to False.
    :type make_pd: bool, optional
    :param explicit: Whether or not to return the approximate matrix explicitly or implicitly in
        form of three update matrices.
    :type explicit: bool
    :return: H, either as np.ndarray (``explicit=='True'``) or represented implicitly via three
        update vectors (also np.ndarrays)
    :rtype: np.ndarray | tuple[np.ndarray, np.ndarray, np.ndarray]
    """
    m = len(y)
    # First form S,Y, and D
    S = s.T
    Y = y.T

    STS = S.T @ H0 @ S
    DK = np.identity(m)
    DK[0, 0] *= np.dot(s[0], y[0])

    # Now use eq. 2.5 to compute R - only have to do this once
    R0 = np.dot(s[0], y[0]).reshape(1, 1)
    R = R0
    for k in range(1, m):

        DK[k, k] *= np.dot(s[k], y[k])

        R = np.concatenate(
            (
                np.concatenate((R0, S[:, :k].T @ Y[:, [k]]), axis=1),
                np.concatenate(
                    (np.zeros((1, R0.shape[1])), np.array([1 / rho[k]]).reshape(1, 1)),
                    axis=1,
                ),
            ),
            axis=0,
        )

        R0 = R

    # Eq 2.22
    L = S.T @ Y - R

    # Now compute term 2 in eq. 5.2
    t2 = scp.linalg.inv(DK + L + L.T - STS)

    # And terms 1 and 2
    t1 = Y - H0 @ S
    t3 = t1.T

    if make_psd:
        # Compute implicit eigen decomposition as shown by Burdakov et al. (2017)
        Q, R = scp.linalg.qr(t1, mode="economic")
        Rit2R = R @ (t2) @ R.T

        # ev holds non-zero eigenvalues of U@D@U.T (e.g., Burdakov et al. (2017))
        ev, P = scp.linalg.eigh(Rit2R, driver="ev")

        # Now find closest PSD.
        fix_idx = (ev + omega) <= 0

        if np.sum(fix_idx) > 0:
            # print("fix VSR1",np.sum(fix_idx),omega,1/omega)
            ev[fix_idx] = -1 * omega  # + np.power(np.finfo(float).eps,0.9)

            # Useful to guarantee that penalized hessian is pd at convergence
            if make_pd:
                eps = np.power(np.finfo(float).eps, 0.5)
                ev[fix_idx] += eps * np.max(ev + omega)

            # while np.any(np.abs(ev) < 1e-7):
            #   ev[np.abs(ev) < 1e-7] += np.power(np.finfo(float).eps,0.7)

            # print(f"implicit ev post shift. abs min: {np.min(np.abs(ev))}, min: {np.min(ev)},
            #   min + 1: {np.min(ev+omega)}, max + 1: {np.max(ev+omega)}",ev[:10]+omega)

            # Now Q @ P @ np.diag(ev) @ P.T @ Q.T = shifted PSD version of U@D@U.T
            # so we can set:
            # shifted_invt2=np.diag(ev)
            # shifted_t2 = np.diag(1/ev)
            # t1 = Q @ P
            # t3 = t1.T = P.T @ Q.T
            t1 = Q @ P
            t3 = P.T @ Q.T
            t2 = np.diag(ev)

    if explicit:
        H = H0 + t1 @ t2 @ t3
        return H
    else:
        return t1, t2, t3


def computeSH(
    yks: np.ndarray,
    sks: np.ndarray,
    nHfd: scp.sparse.csc_array,
    IonHBB: np.ndarray,
    fcols: np.ndarray,
    acols: np.ndarray,
    sample_hessian: bool = True,
    fully_dampened_HBb: bool = False,
    dampen_HBB: float = 1,
    make_pd: bool = False,
    explicit: bool = True,
    form: str = "SR1",
) -> (
    np.ndarray
    | tuple[
        scp.sparse.csc_array,
        scp.sparse.csc_array,
        np.ndarray,
        scp.sparse.csc_array,
        scp.sparse.csc_array,
        scp.sparse.csc_array,
        scp.sparse.csc_array,
        np.ndarray,
        scp.sparse.csc_array,
    ]
):
    """Computes structured approximation to the negative Hessian of the log-likelihood. Combines
    available/partial information about the Hessian with a quasi-Newton approximation to the
    remaining/unavailable (mixed) second partial derivatives.

    First, note that we can always re-order the negative Hessian :math:`\\mathbf{H}` so that

    .. math::
        :nowrap:

        $$
        \\mathbf{H} =
        \\begin{bmatrix}
            \\mathbf{H}_{\\boldsymbol{\\beta}\\boldsymbol{\\beta}} &
            \\mathbf{H}_{\\boldsymbol{\\beta}\\mathbf{b}} \\\\
            \\mathbf{H}_{\\mathbf{b}\\boldsymbol{\\beta}} &
            \\mathbf{H}_{\\mathbf{b}\\mathbf{b}} \\\\
        \\end{bmatrix}
        $$

    Second, consider the Schur complement (:math:`\\mathbf{D}`) of :math:`\\mathbf{H}` with respect
    to :math:`\\mathbf{H}_{\\mathbf{b}\\mathbf{b}}`:

    .. math::
        \\mathbf{D} = \\mathbf{H}_{\\mathbf{b}\\mathbf{b}} -
        \\mathbf{H}_{\\boldsymbol{b}\\boldsymbol{\\beta}}
        (\\mathbf{H}_{\\boldsymbol{\\beta}\\boldsymbol{\\beta}})^{-1}
        \\mathbf{H}_{\\boldsymbol{\\beta}\\mathbf{b}}

    Third, note that (for quadratic negative log-likelihood or Taylor approximation to -llk) we can
    define a structured secant equation for :math:`\\mathbf{H}_{\\mathbf{b}\\mathbf{b}}` using the
    Schur complement:

    .. math::
        \\mathbf{D}\\mathbf{s}_{\\mathbf{b}} =
        \\mathbf{y}_{\\mathbf{b}} -
        \\left(\\left[\\mathbf{0}~ \\mathbf{H}_{\\boldsymbol{b}\\boldsymbol{\\beta}}
        (\\mathbf{H}_{\\boldsymbol{\\beta}\\boldsymbol{\\beta}})^{-1}
        \\mathbf{H}_{\\boldsymbol{\\beta}\\boldsymbol{b}}\\right] +
        \\left[\\mathbf{H}_{\\boldsymbol{b}\\boldsymbol{\\beta}}~\\mathbf{0}\\right]\\right)
        \\mathbf{s}

    Here :math:`\\mathbf{s}` is a step applied to the coefficient vector (i.e.,
    :math:`\\left[\\boldsymbol{\\beta}, \\boldsymbol{b} \\right])`, :math:`\\mathbf{y}` is the
    difference between the gradients of the negative llk after and before taking this step, and
    versions with :math:`\\mathbf{b}` subscripts index elements in :math:`\\boldsymbol{b}`.

    Since the Schur complement meets the structured secant equation, we can estimate it via
    any quasi-Newton routine so that:

    .. math::
        \\hat{\\mathbf{D}} = \\mathbf{I}\\omega +
        \\mathbf{Q}\\boldsymbol{\\Delta}\\mathbf{Q}^\\top

    In which case, considering the structured secant equation above, we have an approximation to
    the lower block of the negative Hessian defined as:

    .. math::
        \\hat{\\mathbf{H}}_{\\mathbf{b}\\mathbf{b}} =
        \\hat{\\mathbf{D}} + \\mathbf{H}_{\\mathbf{b}\\boldsymbol{\\beta}}
        (\\mathbf{H}_{\\boldsymbol{\\beta}\\boldsymbol{\\beta}})^{-1}
        \\mathbf{H}_{\\boldsymbol{\\beta}\\mathbf{b}}

    and the entire negative Hessian:

    .. math::
        :nowrap:

        $$
        \\hat{\\mathbf{H}} =
        \\begin{bmatrix}
            \\mathbf{H}_{\\boldsymbol{\\beta}\\boldsymbol{\\beta}} &
            \\mathbf{H}_{\\boldsymbol{\\beta}\\mathbf{b}} \\\\
            \\mathbf{H}_{\\mathbf{b}\\boldsymbol{\\beta}} &
            \\hat{\\mathbf{H}}_{\\mathbf{b}\\mathbf{b}} \\\\
        \\end{bmatrix}
        $$

    Note, that if we use a quasi Newton update that ensures positive semi-definiteness of the Schur
    complement :math:`\\hat{\\mathbf{D}}`, then :math:`\\hat{\\mathbf{H}}_{\\mathbf{b}\\mathbf{b}}`
    and :math:`\\hat{\\mathbf{H}}` are also guaranteed to be PSD. Finally, we can represent
    :math:`\\hat{\\mathbf{H}}` implicitly via an initial matrix and three updates (consisting of
    8 update matrices):

    .. math::
        :nowrap:

        $$
        \\hat{\\mathbf{H}} =
        \\begin{bmatrix}
            \\mathbf{H}_{\\boldsymbol{\\beta}\\boldsymbol{\\beta}} & \\mathbf{0} \\\\
            \\mathbf{0} & \\mathbf{I}\\omega \\\\
        \\end{bmatrix} +
        \\begin{bmatrix}
            \\mathbf{0} & \\mathbf{0} \\\\
            \\mathbf{0} & \\mathbf{H}_{\\mathbf{b}\\boldsymbol{\\beta}}
            (\\mathbf{H}_{\\boldsymbol{\\beta}\\boldsymbol{\\beta}})^{-1}
            \\mathbf{H}_{\\boldsymbol{\\beta}\\mathbf{b}} \\\\
        \\end{bmatrix} +
        \\begin{bmatrix}
            \\mathbf{0} & \\mathbf{H}_{\\boldsymbol{\\beta}\\mathbf{b}} \\\\
            \\mathbf{H}_{\\mathbf{b}\\boldsymbol{\\beta}} & \\mathbf{0} \\\\
        \\end{bmatrix} +
        \\begin{bmatrix}
            \\mathbf{0} & \\mathbf{0} \\\\
            \\mathbf{0} & \\mathbf{Q}\\boldsymbol{\\Delta}\\mathbf{Q}^\\top \\\\
        \\end{bmatrix}
        $$

    This is returned by this function, so that::

      nH = nH1 + (nH2t1 @ nH2t2 @ nH2t3) + (nH3t1 @ nH3t3) + (nH4t1 @ nH4t2 @ nH4t3)

    Finally, note that :math:`\\hat{\\mathbf{H}} + \\mathbf{S}_{\\boldsymbol{\\lambda}}` can be
    inverted (:math:`\\mathbf{S}_{\\boldsymbol{\\lambda}}` is the total penalty matrix)
    via repeated application of the modified Woodbury identity of Henderson & Searle (1981).

    :param yks: np.ndarray of shape (m,p), where p is the number of coefficients, holding the
        second set ``m`` of update vectors from Byrd, Nocdeal & Schnabel (1992), modified to account
        for structured update.
    :type yks: np.ndarray
    :param sks: np.ndarray of shape (m,p), where p is the number of coefficients, holding the
        first set ``m`` of update vectors from Byrd, Nocdeal & Schnabel (1992), modified to account
        for structured update.
    :type sks: np.ndarray
    :param nHfd: Finite difference approximation matrix which is symmetric sparse matrix with
        ``fcols`` rows and columns set to finite difference approximation of columns of negative
        Hessian of llk
    :type nHfd: scp.sparse.csc_array
    :param IonHBB: The inverse of the top left block of ``nHfd`` after ordering so that column
        (and row) order is ``[fcols, acols]``
    :type IonHBB: np.ndarray
    :param fcols: Array holding indices of columns of negative Hessian approximated via finite
        differencing.
    :type fcols: np.ndarray
    :param acols: Array holding indices of columns of negative Hessian not approximated via FD (
        to be approximated via quasi Newton)
    :type acols: np.ndarray
    :param sample_hessian: Whether the update vectors ``yks`` and ``sks`` have been sampled or
        obtained from the sequential quasi Newton routine used to update the coefficients, defaults
        to True
    :type sample_hessian: bool, optional
    :param fully_dampened_HBb: Whether the top right and lower left rectangular blocks (mixed
        partial derivatives of ``coef[fcols]`` with respect to ``coef[acols]``) have been zeroed,
        defaults to False
    :type fully_dampened_HBb: bool, optional
    :param dampen_HBB: Scaling factor by which to scale the initialidentity matrix used to
        approximate the quasi Newton block, defaults to 1
    :type dampen_HBB: float, optional
    :param make_pd: Whether to enforce numeric positive definiteness, not just PSD, defaults to
        False.
    :type make_pd: bool, optional
    :param explicit: Whether or not to return the approximate matrix explicitly or implicitly in
        form of 8 update matrices, defaults to True
    :type explicit: bool
    :param form: Should the quasi Newton Hessian approximation use a symmetric rank 1 update
        (``qEFSH='SR1'``) that is forced to result in positive semi-definiteness of the
        approximation or the standard bfgs update (``qEFSH='BFGS'``). Defaults to 'SR1'.
    :type form: str
    :return: H, either as np.ndarray (``explicit=='True'``) or represented implicitly via an initial
        matrix and 8 update matrices as defined in the description (i.e.,
        ``nH1, nH2t1, nH2t2, nH2t3, nH3t1, nH3t3, nH4t1, nH4t2, nH4t3``).
    :rtype: np.ndarray | tuple[scp.sparse.csc_array, scp.sparse.csc_array, np.ndarray,
        scp.sparse.csc_array, scp.sparse.csc_array, scp.sparse.csc_array, scp.sparse.csc_array,
        np.ndarray, scp.sparse.csc_array]
    """

    # Combined indices in order so that quasi Newton block is bottom right
    tcols = [*fcols, *acols]

    nF, nA, nT = len(fcols), len(acols), len(tcols)

    # To compute ordered version of hessian we need first permutation matrix P1:
    # len(tcols) * len(tcols) permutation matrix transforming entire nH into order (onH) so that
    # fd approximated block is top left and quasi Newton block is bottom right via
    # onH = P1.T @ nH @ P1
    P1 = scp.sparse.csc_array(
        (
            np.tile(1, nT),
            (tcols, np.arange(nT)),
        ),
        shape=(nT, nT),
    )

    # Now can compute ordered version of finite difference approximated part of nH and extract the
    # relevant blocks
    onHfd = P1.T @ nHfd @ P1

    # Can now extract the block on the diagonal and the off-diagonals holding the mixed derivs
    # onHBB should be PD
    # onHBb might have been dampened.
    onHBB = onHfd[np.ix_(np.arange(nF), np.arange(nF))]
    onHBb = onHfd[np.ix_(np.arange(nF), np.arange(nF, nT))]

    # Now quasi Newton part. First, retain only rows related to b and drop those related to B
    yks = yks[:, acols]
    sks = sks[:, acols]

    # Finally, recompute rho.
    rhos = 1 / np.einsum("ij,ij->i", sks, yks)

    # Can now compute the quasi newton approximation to the Schur complement onH/onHbb (SonHbbqa)
    # and from that the approximation to onHbb.

    # Form initial matrix H0. See Nocedal & Wright, 2004 for sacling factor
    # computed below.
    if len(yks) > 0:
        omega = dampen_HBB * compute_omega(
            yks,
            sks,
            method="mean" if sample_hessian else "NoWr",
        )
    else:
        omega = 1

    H0 = scp.sparse.identity(nA, format="csc") * omega

    # Implicit representation of Schur complement SonHbbqa
    if form == "SR1":
        qat1, qat2, qat3 = computeHSR1(
            sks, yks, rhos, H0, omega, make_psd=True, explicit=False, make_pd=make_pd
        )
    else:
        qat1, qat2, qat3 = computeH(sks, yks, rhos, H0, explicit=False)

    if explicit:
        # At this point we can explicitly form the PSD approximation to nH (conH). First we define
        # P3:
        # len(tcols) * len(acols) permutation matrix to embed quasi Newton approximated block into
        # bottom right corner of ordered negative hessian (onH) via P3 @ onHbbqa @P3.T
        P3 = scp.sparse.csc_array(
            (
                np.tile(1, nA),
                (np.arange(nF, nT), np.arange(nA)),
            ),
            shape=(nT, nA),
        )

        # Now have
        SonHbbqa = H0 + qat1 @ qat2 @ qat3
        onHbbqa = SonHbbqa + onHBb.T @ IonHBB @ onHBb

        # And thus
        onH = onHfd + P3 @ onHbbqa @ P3.T

        # Then we can use P1 to undo the permutation to finally get nH -> this time reversed
        # multiplication order
        nH = P1 @ onH @ P1.T
        return nH

    # Can represent nH implicitly via initial block diagonal matrix and three additive
    # corrections of the form t1 @ t2 @ t3, where t3 is sometimes t1.T

    # Need another permutation matrix for this, P2:
    # len(tcols) * len(acols) permutation matrix transforming quasi Newton approximated block
    # (nHqa) to fit into nH via P2 @ nHqa @ P2.T
    P2 = scp.sparse.csc_array(
        (
            np.tile(1, nA),
            (acols, np.arange(nA)),
        ),
        shape=(nT, nA),
    )

    # Start with block diagonal matrix (before permutation it's block diagonal)
    nH1 = (
        P1
        @ scp.sparse.block_array([[scp.sparse.csc_array(onHBB), None], [None, H0]])
        @ P1.T
    )

    # Now represent first additive update, P2 @ onHBb.T @ IonHBB @ onHBb @ P2.T, implicitly
    nH2t1 = P2 @ onHBb.T
    nH2t2 = IonHBB
    nH2t3 = onHBb @ P2.T

    # For second additive update t2, the central matrix, is the identity so we only need t1 and
    # t3
    if fully_dampened_HBb:
        # Fully dampened case where onHBb is just zeroes
        nH3t1 = None
        nH3t3 = None
    else:
        nH3t1 = P1 @ scp.sparse.block_array(
            [[scp.sparse.eye_array(nF, format="csc"), None], [None, onHBb.T]]
        )
        nH3t3 = (
            scp.sparse.block_array(
                [[None, onHBb], [scp.sparse.eye_array(nF, format="csc"), None]]
            )
            @ P1.T
        )

    # Now final additive update, which is again of the same form as the first, but this time
    # carries the quasi Newton information about the Schur complement
    nH4t1 = scp.sparse.csc_array(P2 @ qat1)
    nH4t2 = qat2
    nH4t3 = scp.sparse.csc_array(qat3 @ P2.T)

    # Now, nH = nH1 + (nH2t1@nH2t2@nH2t3) + (nH3t1@nH3t3) + (nH4t1@nH4t2@nH4t3)
    return nH1, nH2t1, nH2t2, nH2t3, nH3t1, nH3t3, nH4t1, nH4t2, nH4t3


def computeSVPS(
    nH1: scp.sparse.csc_array,
    nH2t1: scp.sparse.csc_array,
    nH2t2: np.ndarray,
    nH2t3: scp.sparse.csc_array,
    nH3t1: scp.sparse.csc_array | None,
    nH3t3: scp.sparse.csc_array | None,
    nH4t1: scp.sparse.csc_array,
    nH4t2: np.ndarray,
    nH4t3: scp.sparse.csc_array,
    S_emb: scp.sparse.csc_array,
    fully_dampened_HBb: bool = False,
    explicit: bool = True,
    n_c: int = 10,
) -> (
    np.ndarray
    | tuple[
        scp.sparse.csc_array,
        scp.sparse.csc_array,
        np.ndarray,
        scp.sparse.csc_array,
        scp.sparse.csc_array,
        np.ndarray,
        scp.sparse.csc_array,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]
):
    """Computes inverse of structured approximation to the negative Hessian of the **penalized
    log-likelihood**.

    Expects matrix of the form returned by :func:`computeSH`, i.e.,:

    .. math::
        :nowrap:

        $$
        \\hat{\\mathbf{H}} =
        \\begin{bmatrix}
            \\mathbf{H}_{\\boldsymbol{\\beta}\\boldsymbol{\\beta}} & \\mathbf{0} \\\\
            \\mathbf{0} & \\mathbf{I}\\omega \\\\
        \\end{bmatrix} +
        \\begin{bmatrix}
            \\mathbf{0} & \\mathbf{0} \\\\
            \\mathbf{0} & \\mathbf{H}_{\\mathbf{b}\\boldsymbol{\\beta}}
            (\\mathbf{H}_{\\boldsymbol{\\beta}\\boldsymbol{\\beta}})^{-1}
            \\mathbf{H}_{\\boldsymbol{\\beta}\\mathbf{b}} \\\\
        \\end{bmatrix} +
        \\begin{bmatrix}
            \\mathbf{0} & \\mathbf{H}_{\\boldsymbol{\\beta}\\mathbf{b}} \\\\
            \\mathbf{H}_{\\mathbf{b}\\boldsymbol{\\beta}} & \\mathbf{0} \\\\
        \\end{bmatrix} +
        \\begin{bmatrix}
            \\mathbf{0} & \\mathbf{0} \\\\
            \\mathbf{0} & \\mathbf{Q}\\boldsymbol{\\Delta}\\mathbf{Q}^\\top \\\\
        \\end{bmatrix}
        $$

    And then computes the inverse :math:`\\mathbf{V}` of
    :math:`\\hat{\\mathbf{H}} + \\mathbf{S}_{\\boldsymbol{\\lambda}}` by applying the modified
    Woodbury identity of Henderson & Searle (1981) three times. Optionally, the second update
    can be skipped in case the off-diagonal derivative blocks
    :math:`\\mathbf{H}_{\\mathbf{b}\\boldsymbol{\\beta}}` contain only zeroes
    (``fully_dampened_HBb is True``).

    Function either returns :math:`\\mathbf{V}` directly, or returns an initial matrix ``V0`` as
    well as nine update matrices, so that::

      V = V0 - (inv1t1 @ inv1t2 @ inv1t3) - (inv2t1 @ inv2t2 @ inv2t3) - (inv3t1 @ inv3t2 @ inv3t3)

    :param nH1: Initial approximation to the negative Hessian matrix
    :type nH1: scp.sparse.csc_array
    :param nH2t1: Hessian update matrix 1
    :type nH2t1: scp.sparse.csc_array
    :param nH2t2: Hessian update matrix 2
    :type nH2t2: np.ndarray
    :param nH2t3: Hessian update matrix 3
    :type nH2t3: scp.sparse.csc_array
    :param nH3t1: Hessian update matrix 4
    :type nH3t1: scp.sparse.csc_array | None
    :param nH3t3: Hessian update matrix 5
    :type nH3t3: scp.sparse.csc_array | None
    :param nH4t1: Hessian update matrix 6
    :type nH4t1: scp.sparse.csc_array
    :param nH4t2: Hessian update matrix 7
    :type nH4t2: np.ndarray
    :param nH4t3: Hessian update matrix 8
    :type nH4t3: scp.sparse.csc_array
    :param S_emb: Total penalty matrix to be added to the negative Hessian
    :type S_emb: scp.sparse.csc_array
    :param fully_dampened_HBb: Whether the top right and lower left rectangular blocks (mixed
        partial derivatives of ``coef[fcols]`` with respect to ``coef[acols]``) have been zeroed,
        defaults to False
    :type fully_dampened_HBb: bool, optional
    :param explicit: Whether or not to return the approximate matrix explicitly or implicitly in
        form of an initial matrix and 9 update matrices, defaults to True
    :type explicit: bool
    :param n_c: Number of cores to use for multi-processing parts. Defaults to 10
    :type n_c: int, optional
    :return: V, either as np.ndarray (``explicit=='True'``) or represented implicitly via an initial
        matrix and 9 update matrices as defined in the description (i.e.,
        ``V0, inv1t1, inv1t2, inv1t3, inv2t1, inv2t2, inv2t3, inv3t1, inv3t2, inv3t3``).
    :rtype: np.ndarray | tuple[scp.sparse.csc_array, scp.sparse.csc_array, np.ndarray,
        scp.sparse.csc_array, scp.sparse.csc_array, np.ndarray, scp.sparse.csc_array,
        np.ndarray, np.ndarray, np.ndarray,]
    """
    # Now seek implicit representation for inverse of nH + S_emb
    # Can be obtained by applying the modified Woodbury Identity of Henderson and Searle 3 times.

    # Start with obtaining inverse of nH1 + S_emb
    Lp, Pr, _ = cpp_cholP(nH1 + S_emb)  # noqa: F405
    LVp0 = compute_Linv(Lp, n_c)  # noqa: F405
    LV0 = apply_eigen_perm(Pr, LVp0)
    V0 = LV0.T @ LV0

    # Now apply Woodbury identity once to get implicit representation of inverse of
    # nH1 + (nH2t1@nH2t2@nH2t3) + S_emb

    invt2 = np.identity(nH2t2.shape[1]) + nH2t2 @ nH2t3 @ V0 @ nH2t1

    U, sv_invt2, VT = scp.linalg.svd(invt2, lapack_driver="gesvd")

    # Now we we can again compute all parts for the modified Woodbury identy to obtain
    # (nH1 + (nH2t1@nH2t2@nH2t3) + S_emb)^{-1}
    inv1t2 = VT.T @ np.diag(1 / sv_invt2) @ U.T
    inv1t1 = V0 @ nH2t1
    inv1t3 = scp.sparse.csc_array(nH2t2 @ (nH2t3 @ V0))

    if fully_dampened_HBb:
        inv2t2 = None
        inv2t1 = None
        inv2t3 = None
    else:
        # Can now form inverse of
        # nH1 + (nH2t1@nH2t2@nH2t3) + (nH3t1@nH3t3) + S_emb
        # From here on things get messy though, so need to be careful never to explicitly evaluate
        # inverse:
        # (nH1 + (nH2t1@nH2t2@nH2t3) + S_emb)^{-1}
        # so need to watch multiplication order.

        V0nH3t1 = scp.sparse.csc_array(
            (V0 @ nH3t1) - (inv1t1 @ inv1t2 @ (inv1t3 @ nH3t1))
        )

        invt2 = np.identity(nH3t1.shape[1]) + nH3t3 @ V0nH3t1

        U, sv_invt2, VT = scp.linalg.svd(invt2, lapack_driver="gesvd")

        # Now we we can again compute all parts for the modified Woodbury identy to obtain
        # (nH1 + (nH2t1@nH2t2@nH2t3) + (nH3t1@nH3t3) + S_emb)^{-1}
        inv2t2 = VT.T @ np.diag(1 / sv_invt2) @ U.T
        inv2t1 = V0nH3t1
        inv2t3 = scp.sparse.csc_array(
            (nH3t3 @ V0) - ((nH3t3 @ inv1t1) @ inv1t2 @ inv1t3)
        )

    # Can now form final inverse of
    # nH1 + (nH2t1@nH2t2@nH2t3) + (nH3t1@nH3t3) + (nH4t1@nH4t2@nH4t3) + S_emb
    # Again need to watch multiplication order to not explicitly evaulate inverse of
    # (nH1 + (nH2t1@nH2t2@nH2t3) + (nH3t1@nH3t3) + S_emb)^{-1}

    if fully_dampened_HBb:
        V0nH4t1 = (V0 @ nH4t1) - (inv1t1 @ inv1t2 @ (inv1t3 @ nH4t1))
    else:
        V0nH4t1 = (
            (V0 @ nH4t1)
            - (inv1t1 @ inv1t2 @ (inv1t3 @ nH4t1))
            - (inv2t1 @ inv2t2 @ (inv2t3 @ nH4t1))
        )

    invt2 = np.identity(nH4t2.shape[1]) + nH4t2 @ nH4t3 @ V0nH4t1

    U, sv_invt2, VT = scp.linalg.svd(invt2, lapack_driver="gesvd")

    # Now we we can again compute all parts for the modified Woodbury identy to obtain
    # (nH1 + (nH2t1@nH2t2@nH2t3) + (nH3t1@nH3t3) + (nH4t1@nH4t2@nH4t3) + S_emb)^{-1}
    inv3t2 = VT.T @ np.diag(1 / sv_invt2) @ U.T
    inv3t1 = V0nH4t1
    if fully_dampened_HBb:
        inv3t3 = nH4t2 @ ((nH4t3 @ V0) - ((nH4t3 @ inv1t1) @ inv1t2 @ inv1t3))
    else:
        inv3t3 = nH4t2 @ (
            (nH4t3 @ V0)
            - ((nH4t3 @ inv1t1) @ inv1t2 @ inv1t3)
            - ((nH4t3 @ inv2t1) @ inv2t2 @ inv2t3)
        )

    # Can form final inverse!
    if explicit:
        V = V0 - (inv1t1 @ inv1t2 @ inv1t3)

        if fully_dampened_HBb is False:
            V -= inv2t1 @ inv2t2 @ inv2t3

        V -= inv3t1 @ inv3t2 @ inv3t3

        return V

    return V0, inv1t1, inv1t2, inv1t3, inv2t1, inv2t2, inv2t3, inv3t1, inv3t2, inv3t3



# flake8: noqa
import mssm
from mssm.models import *
from mssm.src.python.compare import compare_CDL
from mssm.src.python.utils import estimateVp, correct_VB, MVUniformRhoPrior
import numpy as np
import os
import copy
from mssmViz.sim import *
from .defaults import (
    default_compare_test_kwargs,
    default_gamm_test_kwargs,
    default_gammlss_test_kwargs,
    max_atol,
    max_rtol,
    init_coef_gaumlss_tests,
    init_coef_gammals_tests,
    init_penalties_tests_gammlss,
)

mssm.src.python.exp_fam.GAUMLSS.init_coef = init_coef_gaumlss_tests
mssm.src.python.exp_fam.GAMMALS.init_coef = init_coef_gammals_tests
mssm.src.python.exp_fam.GAUMLSS.init_lambda = init_penalties_tests_gammlss
mssm.src.python.exp_fam.GAMMALS.init_lambda = init_penalties_tests_gammlss
mssm.src.python.exp_fam.MULNOMLSS.init_lambda = init_penalties_tests_gammlss

################################################################## Tests ##################################################################


class Test_model_comparisons1:

    # Model comparison and smoothness uncertainty correction tests

    # Simulate some data
    sim_fit_dat = sim3(n=500, scale=2, c=0.1, family=Gaussian(), seed=21)

    # Now fit nested models
    sim_fit_formula = Formula(
        lhs("y"),
        [
            i(),
            f(["x0"], nk=20, rp=1),
            f(["x1"], nk=20, rp=1),
            f(["x2"], nk=20, rp=1),
            f(["x3"], nk=20, rp=1),
        ],
        data=sim_fit_dat,
        print_warn=False,
    )

    sim_fit_model = GAMM(sim_fit_formula, Gaussian())

    test_kwargs_model = copy.deepcopy(default_gamm_test_kwargs)
    test_kwargs_model["exclude_lambda"] = False
    test_kwargs_model["max_outer"] = 100

    sim_fit_model.fit(**test_kwargs_model)

    sim_fit_formula2 = Formula(
        lhs("y"),
        [i(), f(["x1"], nk=20, rp=1), f(["x2"], nk=20, rp=1), f(["x3"], nk=20, rp=1)],
        data=sim_fit_dat,
        print_warn=False,
    )

    sim_fit_model2 = GAMM(sim_fit_formula2, Gaussian())
    sim_fit_model2.fit(**test_kwargs_model)

    prior = MVUniformRhoPrior(b=np.log(1e12))

    # And perform a couple of bias corrected / smoothness uncertainty corrected / or not / comparisons...
    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_V"] = False
    test_kwargs["correct_t1"] = False
    test_kwargs["seed"] = 22
    uncor_result = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ1"
    test_kwargs["seed"] = 22
    test_kwargs["Vp_fidiff"] = True
    cor_result1 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ2"
    test_kwargs["seed"] = 22
    test_kwargs["only_expected_edf"] = True
    test_kwargs["use_importance_weights"] = False
    test_kwargs["Vp_fidiff"] = False
    cor_result2 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 10
    test_kwargs["grid"] = "JJJ2"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = False
    test_kwargs["Vp_fidiff"] = False
    cor_result3 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = True
    test_kwargs["Vp_fidiff"] = False
    cor_result4 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = True
    test_kwargs["Vp_fidiff"] = False
    test_kwargs["prior"] = prior
    cor_result5 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_V"] = False
    test_kwargs["correct_t1"] = True
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    unbiased_uncor_result = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = True
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = True
    test_kwargs["prior"] = prior
    unbiased_cor_result1 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    def test_comp1(self):
        assert round(self.uncor_result["aic_diff"], ndigits=3) == 1.98

    def test_comp2(self):
        assert round(self.cor_result1["aic_diff"], ndigits=3) == 1.981

    def test_comp3(self):
        assert round(self.cor_result2["aic_diff"], ndigits=3) == 1.703

    def test_comp4(self):
        assert round(self.cor_result3["aic_diff"], ndigits=3) == 1.736

    def test_comp5(self):
        assert round(self.cor_result4["aic_diff"], ndigits=3) == 1.981

    def test_comp6(self):
        assert round(self.cor_result5["aic_diff"], ndigits=3) == 2.977

    def test_comp7(self):
        assert round(self.unbiased_uncor_result["aic_diff"], ndigits=3) == 1.98

    def test_comp8(self):
        assert round(self.unbiased_cor_result1["aic_diff"], ndigits=3) == 2.977

    def test_edf1(self):
        assert round(self.unbiased_cor_result1["Res. DOF"], ndigits=3) == 1.48


class Test_model_comparisons2:

    # Model comparison and smoothness uncertainty correction tests

    # Simulate some data
    sim_fit_dat = sim3(n=500, scale=2, c=0.1, family=Gamma(), seed=21)

    # Now fit nested models
    sim_fit_formula = Formula(
        lhs("y"),
        [
            i(),
            f(["x0"], nk=20, rp=1),
            f(["x1"], nk=20, rp=1),
            f(["x2"], nk=20, rp=1),
            f(["x3"], nk=20, rp=1),
        ],
        data=sim_fit_dat,
        print_warn=False,
    )

    sim_fit_model = GAMM(sim_fit_formula, Gamma())

    test_kwargs_model = copy.deepcopy(default_gamm_test_kwargs)
    test_kwargs_model["exclude_lambda"] = False
    test_kwargs_model["max_outer"] = 100

    sim_fit_model.fit(**test_kwargs_model)

    sim_fit_formula2 = Formula(
        lhs("y"),
        [i(), f(["x1"], nk=20, rp=1), f(["x2"], nk=20, rp=1), f(["x3"], nk=20, rp=1)],
        data=sim_fit_dat,
        print_warn=False,
    )

    sim_fit_model2 = GAMM(sim_fit_formula2, Gamma())
    sim_fit_model2.fit(**test_kwargs_model)

    prior = MVUniformRhoPrior(b=np.log(1e12))
    # And perform a couple of bias corrected / smoothness uncertainty corrected / or not / comparisons...
    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_V"] = False
    test_kwargs["correct_t1"] = False
    test_kwargs["seed"] = 22
    uncor_result = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ1"
    test_kwargs["seed"] = 22
    test_kwargs["Vp_fidiff"] = True
    cor_result1 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ2"
    test_kwargs["seed"] = 22
    test_kwargs["only_expected_edf"] = True
    test_kwargs["use_importance_weights"] = False
    test_kwargs["Vp_fidiff"] = False
    cor_result2 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 10
    test_kwargs["grid"] = "JJJ2"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = False
    test_kwargs["Vp_fidiff"] = False
    cor_result3 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = True
    test_kwargs["Vp_fidiff"] = False
    cor_result4 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 10
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = True
    test_kwargs["Vp_fidiff"] = False
    test_kwargs["prior"] = prior
    cor_result5 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_V"] = False
    test_kwargs["correct_t1"] = True
    test_kwargs["n_c"] = 10
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    unbiased_uncor_result = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = True
    test_kwargs["n_c"] = 10
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = True
    test_kwargs["prior"] = prior
    unbiased_cor_result1 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    def test_comp1(self):
        assert round(self.uncor_result["aic_diff"], ndigits=3) == 1.854

    def test_comp2(self):
        assert round(self.cor_result1["aic_diff"], ndigits=3) == 1.871

    def test_comp3(self):
        assert round(self.cor_result2["aic_diff"], ndigits=3) == 2.012

    def test_comp4(self):
        assert round(self.cor_result3["aic_diff"], ndigits=3) == 1.988

    def test_comp5(self):
        assert round(self.cor_result4["aic_diff"], ndigits=3) == 1.873

    def test_comp6(self):
        assert round(self.cor_result5["aic_diff"], ndigits=3) == 2.649

    def test_comp7(self):
        assert round(self.unbiased_uncor_result["aic_diff"], ndigits=3) == 1.854

    def test_comp8(self):
        assert round(self.unbiased_cor_result1["aic_diff"], ndigits=3) == 2.649

    def test_edf1(self):
        assert round(self.unbiased_cor_result1["Res. DOF"], ndigits=3) == 1.374


class Test_model_comparison3_hard:

    # Model comparison and smoothness uncertainty correction tests

    # Simulate some data
    sim_fit_dat = sim3(n=500, scale=2, c=0.1, family=Gamma(), seed=21)

    # Now fit nested models
    sim_fit_formula = Formula(
        lhs("y"),
        [
            i(),
            f(["x0"], nk=20, rp=1),
            f(["x1"], nk=20, rp=1),
            f(["x2"], nk=20, rp=1),
            f(["x3"], nk=20, rp=1),
        ],
        data=sim_fit_dat,
        print_warn=False,
    )

    sim_fit_formula_sd = Formula(lhs("y"), [i()], data=sim_fit_dat, print_warn=False)

    sim_fit_model = GAMMLSS(
        [sim_fit_formula, copy.deepcopy(sim_fit_formula_sd)],
        family=GAMMALS([LOG(), LOGb(-0.01)]),
    )

    test_kwargs_model = copy.deepcopy(default_gammlss_test_kwargs)
    test_kwargs_model["extend_lambda"] = False
    test_kwargs_model["control_lambda"] = 2
    test_kwargs_model["max_outer"] = 200
    test_kwargs_model["max_inner"] = 500

    sim_fit_model.fit(**test_kwargs_model)

    sim_fit_formula2 = Formula(
        lhs("y"),
        [i(), f(["x1"], nk=20, rp=1), f(["x2"], nk=20, rp=1), f(["x3"], nk=20, rp=1)],
        data=sim_fit_dat,
        print_warn=False,
    )

    sim_fit_model2 = GAMMLSS(
        [sim_fit_formula2, copy.deepcopy(sim_fit_formula_sd)],
        family=GAMMALS([LOG(), LOGb(-0.01)]),
    )
    sim_fit_model2.fit(**test_kwargs_model)

    prior = MVUniformRhoPrior(b=np.log(1e12))
    # And perform a couple of bias corrected / smoothness uncertainty corrected / or not / comparisons...
    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_V"] = False
    test_kwargs["correct_t1"] = False
    test_kwargs["seed"] = 22
    uncor_result = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ1"
    test_kwargs["seed"] = 22
    test_kwargs["Vp_fidiff"] = True
    cor_result1 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ2"
    test_kwargs["seed"] = 22
    test_kwargs["only_expected_edf"] = True
    test_kwargs["use_importance_weights"] = False
    test_kwargs["Vp_fidiff"] = False
    cor_result2 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ2"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = False
    test_kwargs["Vp_fidiff"] = False
    cor_result3 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = True
    test_kwargs["Vp_fidiff"] = False
    cor_result4 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = True
    test_kwargs["Vp_fidiff"] = False
    test_kwargs["prior"] = prior
    cor_result5 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_V"] = False
    test_kwargs["correct_t1"] = True
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    unbiased_uncor_result = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = True
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = True
    test_kwargs["prior"] = prior
    unbiased_cor_result1 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    def test_comp1(self):
        np.testing.assert_allclose(
            self.uncor_result["aic_diff"],
            1.912655,
            atol=min(max_atol, 0.002),
            rtol=min(max_rtol, 0.001),
        )

    def test_comp2(self):
        np.testing.assert_allclose(
            self.cor_result1["aic_diff"],
            1.88216,
            atol=min(max_atol, 0.06),
            rtol=min(max_rtol, 0.001),
        )

    def test_comp3(self):
        np.testing.assert_allclose(
            self.cor_result2["aic_diff"],
            2.219442,
            atol=min(max_atol, 0.07),
            rtol=min(max_rtol, 0.001),
        )

    def test_comp4(self):
        np.testing.assert_allclose(
            self.cor_result3["aic_diff"],
            2.197601,
            atol=min(max_atol, 0.06),
            rtol=min(max_rtol, 0.001),
        )

    def test_comp5(self):
        np.testing.assert_allclose(
            self.cor_result4["aic_diff"],
            1.947487,
            atol=min(max_atol, 0.02),
            rtol=min(max_rtol, 0.001),
        )

    def test_comp6(self):
        np.testing.assert_allclose(
            self.cor_result5["aic_diff"],
            2.829273,
            atol=min(max_atol, 0.3),
            rtol=min(max_rtol, 0.001),
        )

    def test_comp7(self):
        np.testing.assert_allclose(
            self.unbiased_uncor_result["aic_diff"],
            1.912655,
            atol=min(max_atol, 0.002),
            rtol=min(max_rtol, 0.001),
        )

    def test_comp8(self):
        np.testing.assert_allclose(
            self.unbiased_cor_result1["aic_diff"],
            2.829273,
            atol=min(max_atol, 0.3),
            rtol=min(max_rtol, 0.001),
        )

    def test_edf1(self):
        np.testing.assert_allclose(
            self.unbiased_cor_result1["Res. DOF"],
            1.432545,
            atol=min(max_atol, 0.002),
            rtol=min(max_rtol, 0.1),
        )


class Test_model_comparisons4:
    sim_fit_dat = sim3(n=500, scale=2, c=0.0, family=Gaussian(), seed=2)

    sim_fit_formula = Formula(
        lhs("y"),
        [i(), f(["x0"]), f(["x1"]), f(["x2"]), f(["x3"])],
        data=sim_fit_dat,
        print_warn=False,
    )

    test_kwargs = copy.deepcopy(default_gamm_test_kwargs)
    test_kwargs["method"] = "QR"
    test_kwargs["control_lambda"] = 2
    test_kwargs["max_inner"] = 1
    test_kwargs["max_outer"] = 200
    test_kwargs["extend_lambda"] = False

    sim_fit_model = GAMM(sim_fit_formula, ScaledT(Identity(), theta=None))
    sim_fit_model.fit(**test_kwargs)

    sim_fit_formula2 = Formula(
        lhs("y"),
        [i(), f(["x1"]), f(["x2"]), f(["x3"])],
        data=sim_fit_dat,
        print_warn=False,
    )

    sim_fit_model2 = GAMM(sim_fit_formula2, ScaledT(Identity(), theta=None))
    sim_fit_model2.fit(**test_kwargs)

    prior = MVUniformRhoPrior(b=np.log(1e12))
    # And perform a couple of bias corrected / smoothness uncertainty corrected / or not / comparisons...
    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_V"] = False
    test_kwargs["correct_t1"] = False
    test_kwargs["seed"] = 22
    uncor_result = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 1
    test_kwargs["grid"] = "JJJ1"
    test_kwargs["seed"] = 22
    test_kwargs["Vp_fidiff"] = True
    cor_result1 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 10
    test_kwargs["grid"] = "JJJ2"
    test_kwargs["seed"] = 22
    test_kwargs["only_expected_edf"] = True
    test_kwargs["use_importance_weights"] = False
    test_kwargs["Vp_fidiff"] = False
    cor_result2 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 10
    test_kwargs["grid"] = "JJJ2"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = False
    test_kwargs["Vp_fidiff"] = False
    cor_result3 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 10
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = True
    test_kwargs["Vp_fidiff"] = False
    cor_result4 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = False
    test_kwargs["n_c"] = 10
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = True
    test_kwargs["Vp_fidiff"] = False
    test_kwargs["prior"] = prior
    cor_result5 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_V"] = False
    test_kwargs["correct_t1"] = True
    test_kwargs["n_c"] = 10
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    unbiased_uncor_result = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    test_kwargs = copy.deepcopy(default_compare_test_kwargs)
    test_kwargs["correct_t1"] = True
    test_kwargs["n_c"] = 10
    test_kwargs["grid"] = "JJJ3"
    test_kwargs["seed"] = 22
    test_kwargs["use_importance_weights"] = True
    test_kwargs["prior"] = prior
    test_kwargs["recompute_H"] = True
    unbiased_cor_result1 = compare_CDL(sim_fit_model, sim_fit_model2, **test_kwargs)

    def test_comp1(self):
        np.testing.assert_allclose(
            self.uncor_result["aic_diff"],
            0.3787553208994723,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_comp2(self):
        np.testing.assert_allclose(
            self.cor_result1["aic_diff"],
            0.4677656167259556,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_comp3(self):
        np.testing.assert_allclose(
            self.cor_result2["aic_diff"],
            0.421643767664591,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_comp4(self):
        np.testing.assert_allclose(
            self.cor_result3["aic_diff"],
            0.421643767664591,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_comp5(self):
        np.testing.assert_allclose(
            self.cor_result4["aic_diff"],
            0.47304417976920377,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_comp6(self):
        np.testing.assert_allclose(
            self.cor_result5["aic_diff"],
            0.47304417976920377,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_comp7(self):
        np.testing.assert_allclose(
            self.unbiased_uncor_result["aic_diff"],
            0.3787553208994723,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_comp8(self):
        np.testing.assert_allclose(
            self.unbiased_cor_result1["aic_diff"],
            0.47304417976920377,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )

    def test_edf1(self):
        np.testing.assert_allclose(
            self.unbiased_cor_result1["Res. DOF"],
            1.0896620088519597,
            atol=min(max_atol, 0),
            rtol=min(max_rtol, 0.1),
        )


class Test_Vb_corrections:
    # Simulate some data - effect of x0 is very very small
    sim_fit_dat = sim3(n=500, scale=2, c=1, family=Gaussian(), seed=21)

    # Now fit nested models
    sim_fit_formula = Formula(
        lhs("y"),
        [
            i(),
            f(["x0"], nk=20, rp=0),
            f(["x1"], nk=20, rp=0),
            f(["x2"], nk=20, rp=0),
            f(["x3"], nk=20, rp=0),
        ],
        data=sim_fit_dat,
        print_warn=False,
    )

    model = GAMM(sim_fit_formula, Gaussian())

    test_kwargs_model = copy.deepcopy(default_gamm_test_kwargs)
    test_kwargs_model["exclude_lambda"] = False
    test_kwargs_model["max_outer"] = 100

    model.fit(**test_kwargs_model)

    # Now fit nested models
    sim_fit_formula2 = Formula(
        lhs("y"),
        [i(), f(["x0"], nk=10, rp=0), f(["x1"], nk=10, rp=0), f(["x2"], nk=10, rp=0)],
        data=sim_fit_dat,
        print_warn=False,
    )

    model2 = GAMM(sim_fit_formula2, Gaussian())
    model2.fit(**test_kwargs_model)

    Vp1, _, _, _, _, _ = estimateVp(model, grid_type="JJJ1", seed=20)

    _, _, Vp2, _, _, total_edf, _, _, _, _ = correct_VB(
        model, grid_type="JJJ1", verbose=True, seed=20, df=2
    )

    _, LI, _, _, _, total_edf2, _, _, _, _ = correct_VB(
        model2, grid_type="JJJ2", verbose=True, seed=20, df=2
    )

    def test_Vp1(self):
        assert np.allclose(
            np.round(self.Vp1, decimals=3),
            np.array(
                [
                    [1.134000e00, -7.000000e-03, 7.000000e-03, -1.050000e-01],
                    [-7.000000e-03, 8.970000e-01, 1.600000e-02, -7.000000e-03],
                    [7.000000e-03, 1.600000e-02, 4.470000e-01, 9.600000e-02],
                    [-1.050000e-01, -7.000000e-03, 9.600000e-02, 5.989569e03],
                ]
            ),
        )

    def test_Vp2(self):
        assert np.allclose(self.Vp1, self.Vp2)

    def test_edf1(self):
        assert np.round(self.total_edf, decimals=3) == 22.15

    def test_edf2(self):
        assert np.round(self.total_edf2, decimals=3) == 16.857
